"use strict";(self.webpackChunkmitosheet=self.webpackChunkmitosheet||[]).push([[568],{1568:(e,l,a)=>{a.r(l),a.d(l,{ExampleModel:()=>s.ExampleModel,ExampleView:()=>s.ExampleView,MODULE_NAME:()=>t.o,MODULE_VERSION:()=>t.Y,default:()=>E});var t=a(8657),s=a(2868);const E=a(4304).default}}]);