define((()=>(()=>{"use strict";var e={126:e=>{var t;window.__webpack_public_path__=(null===(t=document.querySelector("body"))||void 0===t?void 0:t.getAttribute("data-base-url"))+"nbextensions/mitosheet",window.require&&window.require.config({map:{"*":{mitosheet:"nbextensions/mitosheet/index"}}}),e.exports={load_ipython_extension:function(){}}}},t={};return function o(i){var n=t[i];if(void 0!==n)return n.exports;var r=t[i]={exports:{}};return e[i](r,r.exports,o),r.exports}(126)})()));